#include "lcd.h"
#include <avr/pgmspace.h>
#include <stdio.h>
#include <string.h>
#include "ffft.h"

#include <avr/io.h>
#include <util/delay.h>

#ifndef FFT_N
#define FFT_N 128
#endif


#define CHORDSNUM 6
int16_t capture[FFT_N];			/* Wave captureing buffer */
complex_t bfly_buff[FFT_N];		/* FFT buffer */
uint16_t spektrum[FFT_N/2];		/* Spectrum output buffer */

float frequencies[6][6] = {
	{ 82.41, 110.00, 146.83, 196.00, 246.94, 329.63 }, // Standard - EADGBE
	{ 73.42, 110.00, 146.83, 196.00, 246.94, 329.63 }, // Drop D - DADGBE
	{ 77.78, 103.83, 138.59, 185.00, 233.08, 311.13 }, // Half Step Down - Eb Ab Db Gb Bb Eb
	{ 73.42, 98.00,  130.81, 174.61, 220.00, 293.66 }, // Full Step Down - DGCFAD
	{ 65.41, 98.00,  130.81, 174.61, 220.00, 293.66 }, // Drop C - CGCFAD
	{ 69.30, 103.83, 138.59, 185.00, 116.54, 311.13 }  // Drop C# - Db Ab Db Gb Bb Eb
};

char* tunings_text[] = {
    (char*)"EADGBE",
	(char*)"DADGBE",
	(char*)"EbAbDbGbBbEb",
	(char*)"DGCFAD",
	(char*)"CGCFAD",
	(char*)"DbAbDbGbBbEb"
};

int selected_tuning = 0;
int total_num_of_tunings = 6;
float threshold = 0.5; // consider OK when Freq > Target-threshold and Freq < Target+threshold


int ADC_get(void)
{
	int value = 0;
	ADMUX |= (1 << REFS0) | (1 << ADLAR);
	ADCSRA |= (1 << ADSC);
	
	while(ADCSRA & ( 1 << ADSC));
	value = ADC;

	return value;
}

void get_wave (int16_t *buffer, uint16_t count)
{
	ADMUX = 0;
	do {
		*buffer++ = ADC_get();
	} while(--count);

	ADCSRA = 0;
}

void init_buttons() {
	DDRA &= ~((1<<PA1));
	DDRB &= ~((1<<PB0));
	PORTA |= (1<<PA1);
	PORTB |= (1<<PB0);
}

void init_leds() {
	DDRB |= (1 << PB7) | (1 << PB6) | (1 << PB5) | (1 << PB4) | (1 << PB3) | (1 << PB2) | (1 << PB1);
	
	DDRC |= (1 << PC2) | (1 << PC1) | (1 << PC0);
	DDRD |= (1 << PD7) | (1 << PD6) | (1 << PD5) | (1 << PD4) | (1 << PD3) | (1 << PD2) | (1 << PD1) | (1 << PD0);
	
}

void clear_LCD() {
	LCD_printAt(0, "            ");
}


void update_LCD_text() {
	clear_LCD();
	LCD_printAt(0, tunings_text[selected_tuning]);
}

float get_distance(float frequency, int note_index) {
	float distance = frequency - frequencies[selected_tuning][note_index];
	if (distance < 0)
		distance = -distance;
	return distance;	
}


/*
Just a function to turn on i LEDs
*/ 
void print_number(int i) {
	switch(i) {
		case 0:
			PORTD |= (1 << PD6);
			break;
		case 1:
			PORTC |= (1 << PC2);
			PORTD |= (1 << PD5);
			break;
		case 2:
			PORTD |= (1 << PD3);
			PORTD |= (1 << PD2);
			PORTB |= (1 << PB5);
			break;
		case 3:
			PORTC |= (1 << PC2);
			PORTD |= (1 << PD5);
			PORTD |= (1 << PD2);
			PORTB |= (1 << PB5);
			break;
		case 4:
			PORTD |= (1 << PD6);
			PORTD |= (1 << PD7);
			PORTD |= (1 << PD5);
			PORTD |= (1 << PD2);
			PORTB |= (1 << PB5);
			break;
		case 5:
			PORTD |= (1 << PD6);
			PORTD |= (1 << PD7);
			PORTD |= (1 << PD5);
			PORTD |= (1 << PD2);
			PORTB |= (1 << PB5);
			PORTC |= (1 << PC2);
			break;	
	}
	_delay_ms(1000);
}

void check_for_button_press() {
	
	if (!(PINA & (1 << PA1))) {
		// "NEXT" button pressed
		selected_tuning = (selected_tuning + 1) % total_num_of_tunings;
		print_number(selected_tuning);
		update_LCD_text();
	}
	else if(!(PINB & (1 << PB0))) {
		// "PREVIOUS" button pressed
		
		selected_tuning--;
		if (selected_tuning < 0) {
			selected_tuning += total_num_of_tunings;
		}
		print_number(selected_tuning);
		update_LCD_text();
	}
	
	_delay_ms(25);
}


float get_mic_input() {
	
	memset(capture, 0, FFT_N);
	memset(bfly_buff, 0, FFT_N);
	memset(spektrum, 0, FFT_N/2);
	get_wave(capture, FFT_N);
	fft_input(capture, bfly_buff);
	fft_execute(bfly_buff);
	fft_output(bfly_buff, spektrum);
	int max_val = 0;
	int freq = 0;
	int i;
	for (i = 0; i < FFT_N/2; i++) {
		if (spektrum[i] > max_val) {
			max_val = spektrum[i];
			freq = i;
		}
	}
	return freq;
}

int get_closest_note(float frequency) {
	int i;
	float min_distance = 1000;
	int closest_index = 0;
	float dist;
	for (i = 0; i < CHORDSNUM; i++) {
		dist = get_distance(frequency, i);
		if (dist < min_distance) {
			min_distance = dist;
			closest_index = i;
		}
		
	}
	return closest_index;
}


void turn_off_all_LEDs() {
	PORTB = 1;
	PORTC &= ~(1 << PC0) & ~(1 << PC1) & ~(1 << PC2);
	PORTD = 0;
}

void turn_on_all_LEDs() {
	PORTB = 0xff;
	PORTC |= 7;
	PORTD = 0xff;
	
}


void turn_on_left_LED(int note_index) {
	switch(note_index) {
		case 0:
			PORTC |= (1 << PC0);
			break;
		case 1:
			PORTD |= (1 << PD0);
			break;
		case 2:
			PORTD |= (1 << PD3);
			break;
		case 3:
			PORTD |= (1 << PD6);
			break;
		case 4:
			PORTB |= (1 << PB6);
			break;
		case 5:
			PORTB |= (1 << PB3);
			break;
		default:
			break;
	}
}

void turn_on_center_LED(int note_index) {
	switch(note_index) {
		case 0:
			PORTC |= (1 << PC1);
			break;
		case 1:
			PORTD |= (1 << PD1);
			break;
		case 2:
			PORTD |= (1 << PD4);
			break;
		case 3:
			PORTD |= (1 << PD7);
			break;
		case 4:
			PORTB |= (1 << PB5);
			break;
		case 5:
			PORTB |= (1 << PB2);
			break;
		default:
			break;
	}
}

void turn_on_right_LED(int note_index) {
	switch(note_index) {
		case 0:
			PORTC |= (1 << PC2);
			break;
		case 1:
			PORTD |= (1 << PD2);
			break;
		case 2:
			PORTD |= (1 << PD5);
			break;
		case 3:
			PORTB |= (1 << PB7);
			break;
		case 4:
			PORTB |= (1 << PB4);
			break;
		case 5:
			PORTB |= (1 << PB1);
			break;
		default:
			break;
	}
}

void check_tune(float frequency, int note_index) {
	if (frequency < frequencies[selected_tuning][note_index] - threshold) {
		// Too low
		turn_on_left_LED(note_index);
	}
	else if (frequency > frequencies[selected_tuning][note_index] + threshold) {
		// Too high
		turn_on_right_LED(note_index);
	}
	else {
		// Perfect
		turn_on_center_LED(note_index);
	}
}


void ADC_init(void)
{
 	ADMUX = 0;
	ADCSRA = 0;
	ADMUX |= (1 << REFS0);
	ADCSRA |= (1 << ADEN);  // Enable ADC
	ADCSRA |= (1<<ADSC);	//Turn on ADC
}


int main()
{
	
	ADC_init();
    LCD_init();
	init_buttons();
	init_leds();

	float current_freq;
	int current_note;
	LCD_printAt(0, tunings_text[0]);
	
	//turn_on_all_LEDs();
	for(;;) {
		check_for_button_press();
	
		turn_off_all_LEDs();
		current_freq = get_mic_input();
		current_note = get_closest_note(current_freq);
		check_tune(current_freq, current_note);
	}

    return 0;
}